//
// MP3 Sound DecoderV6SE  SX1 Rev.exp.01
// Copyright(C)'2016- Nagoden / Desktop Station / Fujigaya2 / MECY /TRAINO
//
// 2020/4/18 KEIKYU2100 SIEMENS SOUND (cv47 = 5)
//  MP3N18Exx.ino,kasoku.h, VVVF_Sound.h, VVVF_Sound.cpp
// 
// 2020/4/25 Acceleration sound control with MP3 files.
//　MP3N18X01      Nagoden
//　CV66  ブレーキー音開始閾値 24   MP3-> 0304
//  CV67　加速音開始閾値 15         MP3-> 0301
//  CV68　定速音開始閾値 210        MP3-> 0302
//  Cv69　減速音開始閾値 170        MP3-> 0303

#include <Arduino.h>
#include <string.h>
#include <avr/pgmspace.h> 
#include <avr/eeprom.h>
#include "motor_func.h"
#include "motor_ctrl.h"
#include "VVVF_sound.h"

#include <SoftwareSerial.h>
#include "NmraDcc_x.h"
//#include "TimerOne_x.h"
#include "DFPlayer_Mini_Mp3_x.h"

/*************************************************
 * 
 * Declaration of Symbols
 * 
 *************************************************/

//#define DEBUG			//リリースのときはコメントアウトすること
#define MAN_VER_NUMBER  01   /* Release Ver CV07 */
#define MAN_ID_NUMBER 108  /* Manufacture ID */


#define DECODER_ADDRESS 3

//Pin Func  Port
//24  AUX1  A1
//25  F0_F  A2
//26  F0_R  A3
//27  AUX2  A4
//28  AUX5  A5
//30  AUX6  D0
// 9  AUX4  D5
//10  AUX3  D6
#define PIN_AUX1  A1
#define PIN_F0_F  A2
#define PIN_F0_R  A3
#define PIN_AUX2  A4  
//#define PIN_AUX5  A5  //MP3はスピーカ出力になるので使用しない
//#define PIN_AUX6  0   //MP3はスピーカ出力になるので使用しない
#define PIN_AUX4  5
#define PIN_AUX3  6


#define PIN_BUSY 7

#define MOTOR_LIM_MAX 255

#define CV_VSTART 2             //スタート電圧
#define CV_ACCRATIO 3           //加速率
#define CV_DECCRATIO  4         //減速率
#define CV_VMAX 5               //最高電圧
#define CV_DECODERMODE  30      //デコーダモード
#define CV_SOUNDSWITCH 47       //fujigaya2
#define CV_SOUNDSTATE 48        //fujigaya2
#define CV_SOUNDNOTCH 49        //yaasan
#define CV_SOUNDVOLUME 50       //fujigaya2
#define CV_SOUNDRailInt 51      //fujigaya2

#define CV_BEMFcoefficient  54      //BEMF効果度合
#define CV_PI_P 55                  //BEMF Pパラメータ
#define CV_PI_I 56                  //BEMF Iパラメータ
#define CV_BEMFCUTOFF 57            //BEMF カットオフ値
#define CV_MP3_Vol 58             //MP3 Volume

#define CV_BRAKE_Th 66             //MP3 Brake control threshold ブレーキ音開始速度閾値  MP3(304) 2020/04/25 ADD nagoden
#define CV_START_Th 67             //MP3 START control threshold　加速音速度開始閾値　　MP3(301)  2020/04/25 ADD nagoden
#define CV_CONST_Th 68             //MP3 Const control threshold　定速音速度開始閾値　　MP3(302)  2020/04/25 ADD nagoden
#define CV_Deceleration_Th 69      //MP3 Deceleration control threshold 減速音速度開始閾値　MP3(303) 2020/04/25 ADD nagoden

#define CV_FUNC_F0 110           //F00 Function
#define CV_FUNC_F1 111           //F01 Function
#define CV_FUNC_F2 112           //F02 Function
#define CV_FUNC_F3 113           //F03 Function
#define CV_FUNC_F4 114           //F04 Function
#define CV_FUNC_F5 115           //F05 Function
#define CV_FUNC_F6 116           //F06 Function
#define CV_FUNC_F7 117           //F07 Function
#define CV_FUNC_F8 118           //F08 Function
#define CV_FUNC_F9 119           //F09 Function
#define CV_FUNC_F10 120           //F10 Function
#define CV_FUNC_F11 121           //F11 Function
#define CV_FUNC_F12 122           //F12 Function
#define CV_FUNC_F13 123           //F13 Function
#define CV_FUNC_F14 124           //F14 Function
#define CV_FUNC_F15 125           //F15 Function
#define CV_FUNC_F16 126           //F16 Function
#define CV_FUNC_F17 127           //F17 Function
#define CV_FUNC_F18 128           //F18 Function
#define CV_FUNC_F19 129           //F19 Function
#define CV_FUNC_F20 130           //F20 Function
#define CV_FUNC_F21 131           //F21 Function
#define CV_FUNC_F22 132           //F22 Function
#define CV_FUNC_F23 133           //F23 Function
#define CV_FUNC_F24 134           //F24 Function
#define CV_FUNC_F25 135           //F25 Function
#define CV_FUNC_F26 136           //F26 Function
#define CV_FUNC_F27 137           //F27 Function
#define CV_FUNC_F28 138           //F28 Function
#define CV_FUNCSTATE 200

#define WAIT_VOLMP3		20        //MP3 Volumeの通信ウェイト

/*************************************************
 * 
 * Declaration of Classes
 * 
 *************************************************/

NmraDcc	 Dcc;
DCC_MSG	 Packet;

/*************************************************
 * 
 * Declaration of Variables
 * 
 *************************************************/


//Task Schedule
unsigned long gPreviousL1 = 0;
unsigned long gPreviousL2 = 0;
unsigned long gPreviousL3 = 0;
unsigned long gPreviousL4 = 0;
unsigned long gPreviousL5 = 0; //161117 fujigaya2
unsigned long gPreviousL6 = 0; //170419 fujigaya2
unsigned long gIntervalRail = 1000;

//モータ制御関連の変数
uint16_t gSpeedCmd = 0;
uint16_t gSpeedCmd_vvvf = 0;
uint16_t gSpeedCmd_rail = 0;//fujigaya2
uint8_t gPwmDirv = 128;
uint8_t gCV29_Vstart = 0;
uint8_t gCV29Direction = 0;
uint8_t gDecoderMode = 0; //2018/01/01　nagoden

// CV
uint8_t gCV1_SAddr = 3;
uint8_t gCVx_LAddr = 3;
uint8_t gCV2_Vstart = 32;
uint8_t gCV3_AccRatio = 8;
uint8_t gCV4_DecRatio = 8;
uint8_t gCV5_VMAX = 200;
uint8_t gCV30_DecoderMode = 0;  //デコーダモード　0：モータ　1:モータ制御なし　2018/01/01　nagoden

uint8_t gCV47_SoundSwitch = 4;  //2：吊り掛け音   
uint8_t gCV48_SoundState = 2;   //0:レール継ぎ目音なし、1：レール継ぎ目音あり
uint8_t gCV49_SoundNotch = 0;   //0:ノッチオフあり,1:ノッチオフなし
uint8_t gCV50_SoundVolume = 0;  //ボリューム　0-8
uint8_t gCV51_SoundRailInt = 100;    //レール継ぎ目音用最高速　0-100
uint8_t gCV54_BEMFcoefficient = 64; // 0-127 step / Reajust by MECY 2017/10/08
uint8_t gCV55_PI_P = 18;            // Reajust by MECY 2017/10/08
uint8_t gCV56_PI_I = 96;
uint8_t gCV57_BEMF_CutOff = 0;      // 0でBEMF Off 1でBEMF On


//MP3加速・減速音制御閾値 2020/04/25
uint8_t gCV66_BRAKE_Th = 24;             //ブレーキ音開始速度閾値  MP3(304)   2020/04/25 ADD nagoden
uint8_t gCV67_START_Th = 15;             //加速音速度開始閾値　　MP3(301)     2020/04/25 ADD nagoden
uint8_t gCV68_CONST_Th = 210 ;           //定速音速度開始閾値　　MP3(302)     2020/04/25 ADD nagoden
uint8_t gCV69_Deceleration_Th = 170;     //減速音速度開始閾値  　MP3(303)     2020/04/25 ADD nagoden


//MP3 Control
uint8_t gCV58_MP3_Vol = 22;
uint8_t gCVMP3Table[29] = {0};

//Mp3 chip Select
uint8_t Chip_mode = 1;  /*YX6300 "1"   YX5200  "0" */
//In case of YX6300, the folder is "01" and the file name is "001.mp3" to "255.mp3".

//ファンクションの変数
uint8_t gFuncBits[29] = {0};
uint8_t gCVFuncTable[29] = {0};
uint8_t gFuncMp3Bits[29] = {0};
//ブレーキ音用
int dec_speed_state = 1;
//加速音用
int acc_speed_state = 1;

struct CVPair{
  uint16_t	CV;
  uint8_t	Value;
};

CVPair FactoryDefaultCVs [] = {
  {CV_MULTIFUNCTION_PRIMARY_ADDRESS, DECODER_ADDRESS},
  {CV_ACCESSORY_DECODER_ADDRESS_MSB, 0},		//The LSB is set CV 1 in the libraries .h file, which is the regular address location, so by setting the MSB to 0 we tell the library to use the same address as the primary address. 0 DECODER_ADDRESS
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB, 0},	 //XX in the XXYY address
  {CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB, 0},	 //YY in the XXYY address
  {CV_29_CONFIG, 2},	 //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
//  {CV_29_CONFIG, 128},   //Make sure this is 0 or else it will be random based on what is in the eeprom which could caue headaches
  {CV_VSTART, 32},
  {CV_ACCRATIO, 8},
  {CV_DECCRATIO, 8},
  {CV_VMAX, 200},
  {CV_DECODERMODE, 0},
  {CV_SOUNDSWITCH, 4},
  {CV_SOUNDSTATE, 2},
	{CV_SOUNDNOTCH, 0},
  {CV_SOUNDVOLUME, 0},
  {CV_SOUNDRailInt,100},
  {CV_BEMFcoefficient, 64},
  {CV_PI_P, 18},
  {CV_PI_I, 96},
  {CV_BEMFCUTOFF,0},
  {CV_MP3_Vol,22},

  {CV_BRAKE_Th, 24},           //2020/04/25 ADD nagoden
  {CV_START_Th, 15},           //2020/04/25 ADD nagoden
  {CV_CONST_Th, 210},          //2020/04/25 ADD nagoden
  {CV_Deceleration_Th, 170},   //2020/04/25 ADD nagoden
 
  {CV_FUNC_F0,1},  //Light Control
  {CV_FUNC_F0 + 1, 2},  //Room Light Control
  {CV_FUNC_F0 + 2, 11},
  {CV_FUNC_F0 + 3, 10},
  {CV_FUNC_F0 + 4, 10},
  {CV_FUNC_F0 + 5, 10},
  {CV_FUNC_F0 + 6, 10},
  {CV_FUNC_F0 + 7, 10},
  {CV_FUNC_F0 + 8, 12},
  {CV_FUNC_F0 + 9, 0},
  {CV_FUNC_F0 + 10, 0},
  {CV_FUNC_F0 + 11, 0},
  {CV_FUNC_F0 + 12, 0},
  {CV_FUNC_F0 + 13, 0},
  {CV_FUNC_F0 + 14, 0},
  {CV_FUNC_F0 + 15, 0},
  {CV_FUNC_F0 + 16, 0},
  {CV_FUNC_F0 + 17, 0},
  {CV_FUNC_F0 + 18, 0},
  {CV_FUNC_F0 + 19, 0},
  {CV_FUNC_F0 + 20, 0},
  {CV_FUNC_F0 + 21, 0},
  {CV_FUNC_F0 + 22, 0},
  {CV_FUNC_F0 + 23, 0},
  {CV_FUNC_F0 + 24, 0},
  {CV_FUNC_F0 + 25, 0},
  {CV_FUNC_F0 + 26, 0},
  {CV_FUNC_F0 + 27, 0},
  {CV_FUNC_F0 + 28, 0},
	{200 + 0, 0}, // Function status
	{200 + 1, 0},
	{200 + 2, 0},
	{200 + 3, 0},
	{200 + 4, 0},
	{200 + 5, 0},
	{200 + 6, 0},
	{200 + 7, 0},
	{200 + 8, 0},
	{200 + 9, 0},
	{200 + 10, 0},
	{200 + 11, 0},
	{200 + 12, 0},
	{200 + 13, 0},
	{200 + 14, 0},
	{200 + 15, 0},
	{200 + 16, 0},
	{200 + 17, 0},
	{200 + 18, 0},
	{200 + 19, 0},
	{200 + 20, 0},
	{200 + 21, 0},
	{200 + 22, 0},
	{200 + 23, 0},
	{200 + 24, 0},
	{200 + 25, 0},
	{200 + 26, 0},
	{200 + 27, 0},
	{200 + 28, 0}
};

/*************************************************
 * 
 * Declaration of Functions
 * 
 *************************************************/

void(* resetFunc) (void) = 0;  //declare reset function at address 0

uint8_t FactoryDefaultCVIndex = sizeof(FactoryDefaultCVs) / sizeof(CVPair);

//void ControlRoomLight(void);
//void ControlHeadLight(void);

//void Play_VolDown(uint8_t inVolDown);
//void Play_VolUp(uint8_t inVolUp);


/*************************************************
 * 
 * Functions
 * 
 *************************************************/

void notifyCVResetFactoryDefault()
{
	//When anything is writen to CV8 reset to defaults. 

	resetCVToDefault();	 
#ifdef DEBUG
	Serial.println("Resetting...");
#endif
	delay(1000);  //typical CV programming sends the same command multiple times - specially since we dont ACK. so ignore them by delaying
	resetFunc();
}


void resetCVToDefault()
{
	//CVをデフォルトにリセット
#ifdef DEBUG
	Serial.println("CVs being reset to factory defaults");
#endif
	
	for (int j=0; j < FactoryDefaultCVIndex; j++ ){
		Dcc.setCV( FactoryDefaultCVs[j].CV, FactoryDefaultCVs[j].Value);
	}
}



extern void	   notifyCVChange( uint16_t CV, uint8_t Value)
{
   //CVが変更されたときのメッセージ
#ifdef DEBUG
   Serial.print("CV "); 
   Serial.print(CV); 
   Serial.print(" Changed to "); 
   Serial.println(Value, DEC);
#endif
}

void notifyCVAck(void)
{
#ifdef DEBUG
  Serial.println("notifyCVAck");
#endif
  
	MOTOR_Ack();
}

void setup()
{
	
	// ソフトウエアシリアル通信レートセット:
	Serial.begin(9600);
	mp3_set_serial (Serial);	//set softwareSerial for DFPlayer-mini mp3 module 
  //ファンクションの割り当てピン初期化

  pinMode(PIN_F0_F, OUTPUT);
  digitalWrite(PIN_F0_F, 0);
  pinMode(PIN_F0_R, OUTPUT);
  digitalWrite(PIN_F0_R, 0);  
  pinMode(PIN_AUX1, OUTPUT);
  digitalWrite(PIN_AUX1, 0);
  pinMode(PIN_AUX2, OUTPUT);
  digitalWrite(PIN_AUX2, 0);
//  pinMode(PIN_AUX5, OUTPUT);    //MP3はスピーカ出力になるので使用しない
//  digitalWrite(PIN_AUX5, 0);    //MP3はスピーカ出力になるので使用しない
//  pinMode(PIN_AUX6, OUTPUT);    //MP3はスピーカ出力になるので使用しない
//  digitalWrite(PIN_AUX6, 0);    //MP3はスピーカ出力になるので使用しない
   
	pinMode(PIN_BUSY, INPUT);  //Mp3 Busy pin 設定
	digitalWrite(PIN_BUSY, HIGH); //Internal pull up enabled

	//VVVFハード設定
    VVVF_Setup();

	if ( Dcc.getCV(CV_MULTIFUNCTION_PRIMARY_ADDRESS) == 0xFF )
	{
		//if eeprom has 0xFF then assume it needs to be programmed
#ifdef DEBUG
	  Serial.println("CV Defaulting due to blank eeprom");
#endif
	  notifyCVResetFactoryDefault();
   } else{
#ifdef DEBUG
	 Serial.println("CV Not Defaulting");
#endif
   }
  
	// Setup which External Interrupt, the Pin it's associated with that we're using, disable pullup.
	Dcc.pin(0, 2, 0);

	// Call the main DCC Init function to enable the DCC Receiver
	Dcc.init( MAN_ID_NUMBER, MAN_VER_NUMBER,   FLAGS_MY_ADDRESS_ONLY , 0 ); 

	//Reset task
	gPreviousL1 = millis();
	gPreviousL2 = millis();
	gPreviousL3 = millis();
	gPreviousL4 = millis();
	gPreviousL5 = millis();

	//Init CVs
	gCV1_SAddr = Dcc.getCV( CV_MULTIFUNCTION_PRIMARY_ADDRESS ) ;
	gCV2_Vstart = Dcc.getCV( CV_VSTART ) ;
	gCV3_AccRatio = Dcc.getCV( CV_ACCRATIO ) ;
	gCV4_DecRatio = Dcc.getCV( CV_DECCRATIO ) ;
	gCV5_VMAX = Dcc.getCV( CV_VMAX ) ;
	gCVx_LAddr = (Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_MSB ) << 8) + Dcc.getCV( CV_MULTIFUNCTION_EXTENDED_ADDRESS_LSB );
	gCV29_Vstart = Dcc.getCV( CV_29_CONFIG ) ;
  gCV30_DecoderMode = Dcc.getCV( CV_DECODERMODE ) ; // add 2018/01/01 nagoden
	gCV47_SoundSwitch = Dcc.getCV(CV_SOUNDSWITCH);
	gCV48_SoundState = Dcc.getCV(CV_SOUNDSTATE);
	gCV49_SoundNotch = Dcc.getCV(CV_SOUNDNOTCH);
	gCV50_SoundVolume = Dcc.getCV(CV_SOUNDVOLUME);
	gCV51_SoundRailInt = Dcc.getCV(CV_SOUNDRailInt);
	gCV54_BEMFcoefficient =  Dcc.getCV( CV_BEMFcoefficient ) ;
	gCV55_PI_P = Dcc.getCV( CV_PI_P ) ;
	gCV56_PI_I = Dcc.getCV( CV_PI_I ) ;
	gCV57_BEMF_CutOff = Dcc.getCV( CV_BEMFCUTOFF ) ;
	gCV58_MP3_Vol = Dcc.getCV( CV_MP3_Vol ) ;

  gCV66_BRAKE_Th = Dcc.getCV(CV_BRAKE_Th);            //2020/04/25 ADD nagoden
  gCV67_START_Th = Dcc.getCV(CV_START_Th);            //2020/04/25 ADD nagoden
  gCV68_CONST_Th = Dcc.getCV(CV_CONST_Th);             //2020/04/25 ADD nagoden
  gCV69_Deceleration_Th = Dcc.getCV(CV_Deceleration_Th);    //2020/04/25 ADD nagoden

//AUX3,4 pin assignment

  pinMode(PIN_AUX3, OUTPUT);
  digitalWrite(PIN_AUX3, 0);
  pinMode(PIN_AUX4, OUTPUT);
  digitalWrite(PIN_AUX4, 0);

	/* MP3ファンクションテーブル初期化・読み取り */
	
	for( int i = 0; i < 29; i++)
	{
		gCVFuncTable[i] = Dcc.getCV( CV_FUNC_F0 + i ) ;
		
		/* 異常値は0に固定 */
		if( gCVFuncTable[i] > 19)
		{
			gCVFuncTable[i] = 0;
		}
		
	}
	
	//ファンクション状態をロード
	for( int i = 0; i < 29; i++)
	{
		gFuncBits[i] = Dcc.getCV( CV_FUNCSTATE + i ) ;
		
		/* 異常値は0に固定 */
		if( gFuncBits[i] > 1)
		{
			gFuncBits[i] = 0;
		}
	}

	//cv29 Direction Check
	if ( (gCV29_Vstart & 0x01) > 0)
	{
		gCV29Direction = 1;//REVをFWDにする
	}
	else
	{
		gCV29Direction = 0;//FWDをFWDにする
	}

	MOTOR_Init();
	MOTOR_SetCV(2, gCV2_Vstart);
	MOTOR_SetCV(3, gCV3_AccRatio);
	MOTOR_SetCV(4, gCV4_DecRatio);
	MOTOR_SetCV(5, gCV5_VMAX);
	MOTOR_SetCV(54, gCV54_BEMFcoefficient);        //Nagoden Adjust1 2016/06/02
	VVVF_SetCV(3, gCV3_AccRatio);
	VVVF_SetCV(4, gCV4_DecRatio);
	VVVF_SetCV(47, gCV47_SoundSwitch);        // 160917 fujigaya2
	VVVF_SetCV(48, gCV48_SoundState);        // 160917 fujigaya2
	VVVF_SetCV(49, gCV49_SoundNotch);        //161010 Yaasan
	VVVF_SetCV(50, gCV50_SoundVolume);       //161108 fujigaya2
	MOTOR_SetCV(55, gCV55_PI_P);				// MECY Adjust1 2016/04/09
	MOTOR_SetCV(56, gCV56_PI_I);				// MECY Adjust1 2016/04/09
	MOTOR_SetCV(57, gCV57_BEMF_CutOff);
 
    
	//VVVFをCVで初期化
	VVVF_Init();
	
   
#ifdef DEBUG
	Serial.print("CV1(ShortAddr): ");
	Serial.println(gCV1_SAddr);
	Serial.print("CV17/18(LongAddr): ");
	Serial.println(gCVx_LAddr);
	Serial.print("CV2(Vstart): ");
	Serial.println(gCV2_Vstart);
	Serial.print("CV3(AccRatio): ");
	Serial.println(gCV3_AccRatio);
	Serial.print("CV4(DecRatio): ");
	Serial.println(gCV4_DecRatio);
	Serial.print("CV5(VoltageMAX): ");
	Serial.println(gCV5_VMAX);

	Serial.print(F("CV47(SoundSwitch): "));
	Serial.println(gCV47_SoundSwitch);
	Serial.print(F("CV48(SoundState): "));
	Serial.println(gCV48_SoundState);
  
	Serial.print("CV54 (BEMF coefficient): ");
	Serial.println(gCV54_BEMFcoefficient);
	Serial.print("CV55 (PI P Gain): ");
	Serial.println(gCV55_PI_P);
	Serial.print("CV56 (PI I gain): ");
	Serial.println(gCV56_PI_I);
	Serial.print("CV57 (BEMF Cut Off): ");
	Serial.println(gCV57_BEMF_CutOff);
	Serial.print("CV58 (MP3 Volume): ");
	Serial.println(gCV58_MP3_Vol);
  Serial.print("gCV66_BRAKE_Th: ");
  Serial.println (gCV66_BRAKE_Th);
  Serial.print("gCV67_START_Th: ");
  Serial.println(gCV67_START_Th);
  Serial.print("gCV68_CONST_Th: ");
  Serial.println(gCV68_CONST_Th);
  Serial.print("gCV69_Deceleration_Th: ");
  Serial.println(gCV69_Deceleration_Th);	
	
	Serial.println("Ready");
#endif

}

void loop()
{
	
	// You MUST call the NmraDcc.process() method frequently from the Arduino loop() function for correct library operation
	Dcc.process();
	
	if( (millis() - gPreviousL1) >= 10)
	{
    if (gCV30_DecoderMode != 1 )  //2018/01/01 Nagoden
    {
  		//Speed detection
	  	MOTOR_Sensor();
    }
		//Reset task
		gPreviousL1 = millis();

	}	

	if( (millis() - gPreviousL2) >= 33)
	{
		//VVVF control 京急とそれ以外で分ける
	

    // // 2020/4/18 KEIKYU2100 / TRAIO
		if(gCV47_SoundSwitch==5)
      KQ2100_VVVF_Cont(gSpeedCmd_vvvf, gFuncBits[3]);
    else
   	  VVVF_Cont_Normal(gSpeedCmd_vvvf, gFuncBits[3]);

    //Interval_Rail [ms]
    //255で正規化されているはずなのでとりあえず最高速100km/h固定として、
    //int kmph = gSpeedCmd_vvvf * 100 / 255;
    //int mps = kmph * 1000 / 3600;
    //int t= 25 / mps;
		gSpeedCmd_rail = (gSpeedCmd_vvvf * gCV51_SoundRailInt) / 100L;  
		if(gSpeedCmd_rail != 0)
		{
			gIntervalRail =  (255L * 900L) / (unsigned long)gSpeedCmd_rail - 200;
		}
		//Reset task
		gPreviousL2 = millis();

	} 
	
	if( (millis() - gPreviousL3) >= 100)
	{
    if (gCV30_DecoderMode != 1 ) //2018/01/01 Nagoden
    {
		  //Motor drive control
		  MOTOR_Main(gSpeedCmd, gPwmDirv);
    }
		//Reset task
		gPreviousL3 = millis();   
	}
	
	
	if( (millis() - gPreviousL4) >= 250)
	{	
		//Reset task
		gPreviousL4 = millis();
	}

  if( (millis() - gPreviousL5) >= gIntervalRail)
  { 
    //Reset task
    gPreviousL5 = millis();
    //Rail Sound
    if((gCV48_SoundState & 0x01) != 0)
    {
      if(gSpeedCmd_rail != 0)
      {
        Play_RailSound(gSpeedCmd_rail);
      }
    }
  }
// MP3V6 加速・定速・減速・ブレーキ音制御
  if( (millis() - gPreviousL6) >= 100)
  {
    if((gCV48_SoundState & 0x02) != 0)//170805 fujigaya2
    {
      //gSpeedCmd_vvvfは255最大
      if ( gSpeedCmd_vvvf > gCV66_BRAKE_Th ) // ブレーキ音発生イベント検出
      {
        dec_speed_state = 1;
      }
      if ( gSpeedCmd_vvvf <= gCV66_BRAKE_Th && dec_speed_state == 1)
      {
        dec_speed_state = 2;
        mp3_set_volume (gCV58_MP3_Vol);
        delay (WAIT_VOLMP3);
        mp3_play (204, Chip_mode); //　ブレーキ音 
      }
      if ( gSpeedCmd_vvvf == 0)
      {
           acc_speed_state = 1;  //スピードが０になったら加速音可とする
      }
     //加速音
     if ( gSpeedCmd_vvvf >= gCV67_START_Th && acc_speed_state == 1) // 加速音開始条件
     {
        acc_speed_state = 2;
        mp3_set_volume (gCV58_MP3_Vol);
        delay (WAIT_VOLMP3);
        mp3_play (201, Chip_mode); //　加速音 
     }
     if ((gSpeedCmd_vvvf >= gCV68_CONST_Th) && ((acc_speed_state == 2) || (acc_speed_state == 3 )) ) // 定速音開始条件
     {
        acc_speed_state = 4;
        mp3_set_volume (gCV58_MP3_Vol);
        delay (WAIT_VOLMP3);
        mp3_play (202, Chip_mode);  //　定速音 
        delay (200);
        mp3_single_loop (true);
     }
     if ( gSpeedCmd_vvvf >= gCV69_Deceleration_Th && acc_speed_state == 2) // 減速開始条件(定速なし）
     {
       acc_speed_state =3;
     }

     if (( gSpeedCmd_vvvf < gCV69_Deceleration_Th ) && ((acc_speed_state == 3) || (acc_speed_state == 4 )))    // 減速音開始条件
     {
       acc_speed_state =2;
        mp3_set_volume (gCV58_MP3_Vol);
        delay (WAIT_VOLMP3);
        mp3_play (203, Chip_mode); //　減速音
     }
    }
    //Reset task
    gPreviousL6 = millis();
  }
}



//DCC速度信号の受信によるイベント
extern void notifyDccSpeed( uint16_t Addr, DCC_ADDR_TYPE AddrType, uint8_t Speed, DCC_DIRECTION Dir, DCC_SPEED_STEPS SpeedSteps )
{
	uint16_t aSpeedRef = 0;
	//速度値の正規化(255を100%とする処理)
	if( Speed >= 1)
	{
		aSpeedRef = ((Speed - 1) * 255) / SpeedSteps;
	}
	else
	{
		//緊急停止信号受信時の処理 //Nagoden comment 2016/06/11
#ifdef DEBUG
		Serial.println("***** Emagency STOP **** ");
#endif
		aSpeedRef = 0;
	}

  //VVVF用
  gSpeedCmd_vvvf = aSpeedRef; 
  aSpeedRef = aSpeedRef * gCV5_VMAX / 255; 
	
	gSpeedCmd = aSpeedRef;
  gPwmDirv = Dir;
  if ( gCV29Direction > 0)
  {
     if ( Dir == DCC_DIR_FWD)
     {
          gPwmDirv = 128;
     }
    else
    {
          gPwmDirv = 0;     
    }     
  }
  else
  {
     if ( Dir == DCC_DIR_FWD)
     {
          gPwmDirv = 0;
     }
    else
    {
         gPwmDirv =128;     
    }     
  }

#ifdef DEBUG
	// デバッグメッセージ
//	Serial.print("Speed - ADR: ");/
//	Serial.print(Addr);
//	Serial.print(", SPD: ");
//	Serial.print(Speed);
//	Serial.print(", DIR: ");
//	Serial.print(Dir);
//	Serial.print(", aSpeedRef: ");
//	Serial.println(aSpeedRef);
#endif

}

//ファンクション信号受信のイベント
extern void notifyDccFunc( uint16_t Addr, DCC_ADDR_TYPE AddrType, FN_GROUP FuncGrp, uint8_t FuncState)
{
  switch(FuncGrp)
  {
    case FN_0_4:
//      Function F00
         if  ((FuncState & FN_BIT_00) == 16)
         {
            if (gFuncBits[0] == 0)
            {
//               Dcc.setCV(CV_FUNCSTATE + 0, 1);
            }
             gFuncBits[0] = 1;
             exec_function (gFuncBits[0], gCVFuncTable[0] ,0);
         }    
         else
         {
            if (gFuncBits[0] == 1) 
            {
              gFuncBits[0] = 0;
              exec_function ( gFuncBits[0], gCVFuncTable[0], 0 );
  //            Dcc.setCV(CV_FUNCSTATE + 0, 0);
            }
         }
             
//      Function F01
         if  ((FuncState & FN_BIT_01) == 1){
            if (gFuncBits[1] == 0)
            {              
              gFuncBits[1] = 1;
              exec_function ( gFuncBits[1], gCVFuncTable[1], 1 );
//              Dcc.setCV(CV_FUNCSTATE + 1, 1);
            }
         }    
         else
         {
            if (gFuncBits[1] == 1) 
            {
              gFuncBits[1] = 0;
              exec_function ( gFuncBits[1], gCVFuncTable[1], 1 );
//              Dcc.setCV(CV_FUNCSTATE + 1, 0);
            }
         }
//      Function F02
         if  ((FuncState & FN_BIT_02) == 2)
         {
           gFuncBits[2] = 1;
           exec_function (gFuncBits[2] , gCVFuncTable[2], 2 );
         }    
         else 
          {
           gFuncBits[2] = 0;
           exec_function ( gFuncBits[2], gCVFuncTable[2], 2 ); 
         }
//      Function F03
         if  ((FuncState & FN_BIT_03) == 4)  
         {
            if( gCV49_SoundNotch != 2){
               gFuncBits[3] = 1;
              exec_function ( gFuncBits[3], gCVFuncTable[3], 3 );
            }
            else {
               gFuncBits[3] = 1;
            }
         }
         else {
            if( gCV49_SoundNotch != 2){
               gFuncBits[3] = 0;
               exec_function ( gFuncBits[3], gCVFuncTable[3], 3 );
            }
            else {
               gFuncBits[3] = 0;
            }  
         }
       
         
//      Function F04 
         if ((FuncState & FN_BIT_04) == 8)
         {
           gFuncBits[4] = 1;
           exec_function ( gFuncBits[4], gCVFuncTable[4], 4 );
         }
         else
         {
           gFuncBits[4] = 0;
           exec_function ( gFuncBits[4], gCVFuncTable[4], 4 );
         } 
      break;
    case FN_5_8:
//      Function F05    
         if  ((FuncState & FN_BIT_05) == 1)
         {
           gFuncBits[5] = 1;
           exec_function ( gFuncBits[5], gCVFuncTable[5], 5 );
         }       
         else
         {
           gFuncBits[5] = 0;
           exec_function ( gFuncBits[5], gCVFuncTable[5], 5 );
         } 
//      Function F06    
         if  ((FuncState & FN_BIT_06) == 2) 
         {
            gFuncBits[6] = 1;
            exec_function ( gFuncBits[6], gCVFuncTable[6], 6 );
         }    
         else 
         {
            gFuncBits[6] = 0;
            exec_function ( gFuncBits[6], gCVFuncTable[6], 6 );
         }
//      Function F07    
         if  ((FuncState & FN_BIT_07) == 4) 
         {
            gFuncBits[7] = 1;
            exec_function ( gFuncBits[7], gCVFuncTable[7], 7 );
         }    
         else 
         {
            gFuncBits[7] = 0;
            exec_function ( gFuncBits[7], gCVFuncTable[7], 7 );
         }
//      Function F08    
         if  ((FuncState & FN_BIT_08) == 8)
         {
            gFuncBits[8] = 1;
            exec_function ( gFuncBits[8], gCVFuncTable[8], 8 );
         }    
         else 
         {
            gFuncBits[8] = 0;
            exec_function ( gFuncBits[8], gCVFuncTable[8], 8 );
          }
      break;
    case FN_9_12:
//      Function F09    
         if  ((FuncState & FN_BIT_09) == 1) 
         {
            gFuncBits[9] = 1;
            exec_function ( gFuncBits[9], gCVFuncTable[9], 9 );
         }    
         else 
         {
            gFuncBits[9] = 0;
            exec_function ( gFuncBits[9], gCVFuncTable[9], 9 );
         }
//      Function F10    
         if  ((FuncState & FN_BIT_10) == 2) 
         {
            gFuncBits[10] = 1;
            exec_function ( gFuncBits[10], gCVFuncTable[10], 10 );
         }    
         else 
         {
            gFuncBits[10] = 0;
            exec_function ( gFuncBits[10], gCVFuncTable[10], 10 );
         }
//      Function F11    
         if  ((FuncState & FN_BIT_11) == 4) 
         {
            gFuncBits[11] = 1;
            exec_function ( gFuncBits[11], gCVFuncTable[11], 11 );
         }    
         else 
         {
            gFuncBits[11] = 0;
            exec_function ( gFuncBits[11], gCVFuncTable[11], 11 );
         }

//      Function F12    
         if  ((FuncState & FN_BIT_12) == 8) 
         {
            gFuncBits[12] = 1;
            exec_function ( gFuncBits[12], gCVFuncTable[12], 12 );
         }    
         else 
         {
            gFuncBits[12] = 0;
            exec_function ( gFuncBits[12], gCVFuncTable[12], 12 );
         }

      break;

    case FN_13_20:
//      Function F13
         if  ((FuncState & FN_BIT_13) == 1) 
         {
            gFuncBits[13] = 1;
            exec_function ( gFuncBits[13], gCVFuncTable[13], 13 );
         }    
         else 
         {
            gFuncBits[13] = 0;
            exec_function ( gFuncBits[13], gCVFuncTable[13], 13 );
         }
//      Function F14    
         if  ((FuncState & FN_BIT_14) == 2) 
         {
            gFuncBits[14] = 1;
            exec_function ( gFuncBits[14], gCVFuncTable[14], 14 );
         }    
         else 
         {
            gFuncBits[14] = 0;
            exec_function ( gFuncBits[14], gCVFuncTable[14], 14 );
         }
//      Function F15    
         if  ((FuncState & FN_BIT_15) == 4) 
         {
            gFuncBits[15] = 1;
            exec_function ( gFuncBits[15], gCVFuncTable[15], 15 );
         }    
         else 
         {
            gFuncBits[15] = 0;
            exec_function ( gFuncBits[15], gCVFuncTable[15], 15 );
         }
//      Function F16    
         if  ((FuncState & FN_BIT_16) == 8) 
         {
            gFuncBits[16] = 1;
            exec_function ( gFuncBits[16], gCVFuncTable[16], 16 );
         }    
         else 
         {
            gFuncBits[16] = 0;
            exec_function ( gFuncBits[16], gCVFuncTable[16], 16 );
         }
//      Function F17    
         if  ((FuncState & FN_BIT_17) == 16) 
         {
            gFuncBits[17] = 1;
            exec_function ( gFuncBits[17], gCVFuncTable[17], 17 );
         }    
         else 
         {
            gFuncBits[17] = 0;
            exec_function ( gFuncBits[17], gCVFuncTable[17], 17 );
         }
//      Function F18    
         if  ((FuncState & FN_BIT_18) == 32) 
         {
            gFuncBits[18] = 1;
            exec_function ( gFuncBits[18], gCVFuncTable[18], 18 );
         }    
         else 
         {
            gFuncBits[18] = 0;
            exec_function ( gFuncBits[18], gCVFuncTable[18], 18 );
         }
//      Function F19    
         if  ((FuncState & FN_BIT_19) == 64) 
         {
            gFuncBits[19] = 1;
            exec_function ( gFuncBits[19], gCVFuncTable[19], 19 );
         }    
         else 
         {
            gFuncBits[19] = 0;
            exec_function ( gFuncBits[19], gCVFuncTable[19], 19 );
         }

 //      Function F20    
        if  ((FuncState & FN_BIT_20) == 128) 
        {
            gFuncBits[20] = 1;
            exec_function ( gFuncBits[20], gCVFuncTable[20], 20 );
         }    
         else 
         {
            gFuncBits[20] = 0;
            exec_function ( gFuncBits[20], gCVFuncTable[20], 20 );
         }
      break;
      
    case FN_21_28:
//      Function F21    

         if  ((FuncState & FN_BIT_21) == 1) 
         {
            gFuncBits[21] = 1;
            exec_function ( gFuncBits[21], gCVFuncTable[21], 21 );
         }    
         else 
         {
            gFuncBits[21] = 0;
            exec_function ( gFuncBits[21], gCVFuncTable[21], 21 );
         }
//      Function F22    
         if  ((FuncState & FN_BIT_22) == 2) 
         {
            gFuncBits[22] = 1;
            exec_function ( gFuncBits[22], gCVFuncTable[22], 22 );
         }    
         else 
         {
            gFuncBits[22] = 0;
            exec_function ( gFuncBits[22], gCVFuncTable[22], 22 );
         }
//      Function F23    
         if  ((FuncState & FN_BIT_23) == 4) 
         {
            gFuncBits[23] = 1;
            exec_function ( gFuncBits[23], gCVFuncTable[23], 23 );
         }    
         else 
         {
            gFuncBits[23] = 0;
            exec_function ( gFuncBits[23], gCVFuncTable[23], 23 );
         }
//      Function F24    
         if  ((FuncState & FN_BIT_24) == 8) 
         {
            gFuncBits[24] = 1;
            exec_function ( gFuncBits[24], gCVFuncTable[24], 24 );
         }    
         else 
         {
            gFuncBits[24] = 0;
            exec_function ( gFuncBits[24], gCVFuncTable[24], 24 );

         }
//      Function F25    
         if  ((FuncState & FN_BIT_25) == 16) 
         {
            gFuncBits[25] = 1;
            exec_function ( gFuncBits[25], gCVFuncTable[25], 25 );
         }    
         else 
         {
            gFuncBits[25] = 0;
            exec_function ( gFuncBits[25], gCVFuncTable[25], 25 );
         }
//      Function F26    
         if  ((FuncState & FN_BIT_26) == 32) 
         {
            gFuncBits[26] = 1;
            exec_function ( gFuncBits[26], gCVFuncTable[26], 26 );
         }    
         else 
         {
            gFuncBits[26] = 0;
            exec_function ( gFuncBits[26], gCVFuncTable[26], 26 );
         }
//      Function F27    
         if  ((FuncState & FN_BIT_27) == 64) 
         {
            gFuncBits[27] = 1;
            exec_function ( gFuncBits[27], gCVFuncTable[27], 27 );
         }    
         else 
         {
            gFuncBits[27] = 0;
            exec_function ( gFuncBits[27], gCVFuncTable[27], 27 );
         }
 //      Function F28    
         if  ((FuncState & FN_BIT_28) == 128) 
         {
            gFuncBits[28] = 1;
            exec_function ( gFuncBits[28], gCVFuncTable[28], 28 );
         }    
         else 
         {
            gFuncBits[28] = 0;
            exec_function ( gFuncBits[28], gCVFuncTable[28], 28 );
         }
     break;
  }
  
}

//exec_function   
// FuncState　OFF=0 ON=1
// FuncCV     0:OFF 1:F0  2:AUX1  3:AUX2  4:AUX3  5:AUX4  6:AUX5  7:AUX6  8:AUX3EXT 9:AUX4EXT
//            10: MP3 ON/OFFの切替時のみ単独再生 11:MP3ON時のみ単独再生・OFFの時停止
//            12: MP3 ON時リピート再生・OFF時停止   
// FuncNo     FunctionNumber 0-28        
//   FuncCVはCV111から138までの値
void exec_function ( int FuncStat, int FuncCV ,int FuncNo )
{
 switch( FuncCV )
    {
         case 0:  // Function Off
          break ;     // 0は何もしない

         case 1: // F0の処理
             ControlHeadLight();
          break ;
         case 2: // AUX1の処理
          if( FuncStat == 0) {
            digitalWrite(PIN_AUX1, 0);
          }
          else  {
            digitalWrite(PIN_AUX1, 1);
          }
          break ;
        case 3: // AUX2の処理
          if( FuncStat == 0) {
            digitalWrite(PIN_AUX2, 0);
          }
          else  {
            digitalWrite(PIN_AUX2, 1);
          }
          break ;
        case 4: // AUX3の処理
          if( FuncStat == 0) {
            digitalWrite(PIN_AUX3, 0);
          }
          else  {
            digitalWrite(PIN_AUX3, 1);
          }
          break ;
        case 5: // AUX4の処理   
          if( FuncStat == 0) {
            digitalWrite(PIN_AUX4, 0);
          } 
          else  {
            digitalWrite(PIN_AUX4, 1); 
          }  
          break ;
//        case 6: // AUX5の処理               //MP3はスピーカ出力になるので使用しない
//          if( FuncStat == 0) {              //MP3はスピーカ出力になるので使用しない
//            digitalWrite(PIN_AUX5, 0);      //MP3はスピーカ出力になるので使用しない
//          }                                 //MP3はスピーカ出力になるので使用しない
//          else  {                           //MP3はスピーカ出力になるので使用しない
//            digitalWrite(PIN_AUX5, 1);      //MP3はスピーカ出力になるので使用しない
//          }                                 //MP3はスピーカ出力になるので使用しない
//          break ;                           //MP3はスピーカ出力になるので使用しない
//        case 7: // AUX6の処理               //MP3はスピーカ出力になるので使用しない
//          if( FuncStat == 0) {              //MP3はスピーカ出力になるので使用しない
//            digitalWrite(PIN_AUX6, 0);      //MP3はスピーカ出力になるので使用しない
//          }                                 //MP3はスピーカ出力になるので使用しない
//          else  {                           //MP3はスピーカ出力になるので使用しない
//            digitalWrite(PIN_AUX6, 1);      //MP3はスピーカ出力になるので使用しない
//          }                                 //MP3はスピーカ出力になるので使用しない
//          break ;                           //MP3はスピーカ出力になるので使用しない
        case 8: // AUX3の拡張処理
          break ;
        case 9: // AUX4の拡張処理
          break ;
        case 10: // ON/OFFの切替時のみ単独再生
           if(FuncStat == 1){ 
               PlayMP3_P(FuncNo,FuncCV);
           }
           else {
               PlayMP3_N(FuncNo,FuncCV);           
           }
          break ;
        case 11: // ON時のみ単独再生・OFFの時停止
            if(FuncStat == 1){ 
               PlayMP3_P(FuncNo,FuncCV);
           }
           else {
               PlayMP3_N(FuncNo,FuncCV);           
           }
          break ;
 
        case 12: //ON時リピート再生・OFF時停止
           if(FuncStat == 1){ 
               PlayMP3_P(FuncNo,FuncCV);
           }
           else {
               PlayMP3_N(FuncNo,FuncCV);           
           }
          break ;
        }
}

void ControlHeadLight(void)
{
  if(gFuncBits[0] == 0)
  {
    digitalWrite(PIN_F0_F, 0);
    digitalWrite(PIN_F0_R, 0);  
    return;
  }
  if( gPwmDirv == 0)
  {
    //FWD
      digitalWrite(PIN_F0_F, 1);
      digitalWrite(PIN_F0_R, 0);
      return;
  }
  else
  {
    //REV
      digitalWrite(PIN_F0_F, 0);
      digitalWrite(PIN_F0_R, 1);  
      return;
   
  }
}

void PlayMP3_P(uint8_t inMusicNo, uint8_t inFuncCV)
{
  PlayMP3_PosiEdge( inMusicNo, inFuncCV);
  //Dcc.setCV(CV_FUNCSTATE + inMusicNo, 1);
}

void PlayMP3_N(uint8_t inMusicNo, uint8_t inFuncCV)
{
  PlayMP3_NegaEdge( inMusicNo, inFuncCV);
// Dcc.setCV(CV_FUNCSTATE + inMusicNo, 0);
}

void PlayMP3_PosiEdge(uint8_t inMusicNo, uint8_t inModeCV)
{
  /* 0: ON/OFFのエッジで再生 */
  /* 1: ONレベルで再生(リピート無し) */
  /* 2: ONレベルで再生(リピート有り) */
   
  if ( gFuncMp3Bits[inMusicNo] == 1)  {
     gFuncMp3Bits[inMusicNo] = 0;
     mp3_set_volume (gCV58_MP3_Vol);
     delay (WAIT_VOLMP3);
    if ((inModeCV == 10) || (inModeCV == 11)) {
        mp3_play (inMusicNo,Chip_mode);
    }
    else {                        //inModeCV = 12の場合のリピート
        mp3_play (inMusicNo,Chip_mode);
       delay (200);
       mp3_single_loop (true);
       delay (WAIT_VOLMP3);
    }
  }
  else {
    /* 何もしない */
  }
}

void PlayMP3_NegaEdge(uint8_t inMusicNo, uint8_t inModeCV)
{
  if ( gFuncMp3Bits[inMusicNo] != 1)
  {
    gFuncMp3Bits[inMusicNo] = 1 ;
    if ((inModeCV == 12) || (inModeCV == 11)) {
      /* 11: ONレベルで再生(リピート無し) */
      /* 12: ONレベルで再生(リピート有り) */
          mp3_single_loop (false);
          delay (WAIT_VOLMP3);
          mp3_stop () ;
          delay (WAIT_VOLMP3);
    }
    else if (inModeCV == 10) {
      /* 10: ON/OFFのエッジで再生 */
      mp3_set_volume (gCV58_MP3_Vol);
      delay (WAIT_VOLMP3);
      mp3_play (inMusicNo,Chip_mode);
    }
  }
  else
  {
    //何もしない
  }
}

//YX6300用に変更02020/08/01
void Play_RailSound(uint16_t t_spd)
{
	//255で正規化されているはずなのでとりあえず最高速100km/h固定として、
	int t_switch = t_spd / 26;//0 - 9
	PlayMP3_Simple(101 + t_switch ,Chip_mode);// 101 - 110
}
void PlayMP3_Simple(uint8_t num ,uint8_t Chip_mode)
{
  //Busyが立っている＝ほかの音がなっているときは鳴らさない。
  if(digitalRead(PIN_BUSY) == HIGH)
  {
    //MP3 Volume Set
     mp3_play (num,Chip_mode);
  }
}
